codeunit 50100 SubscribeToAnveoAppEvents
{
    var
        Mgt: Codeunit "ACF Management";
        Configuration: Record "ACF Configuration";
        Text002: Label '%1 %2 already posted or deleted.';
        Text007: Label 'The approval request %1 %2 cannot be canceled because the order has already been released. ';
        Text009: Label 'Your data has been succesfully processed.';
        Text010: Label 'Your data has been processed.';
        Error001: Label 'Could not create %1: %2';

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'OnSyncRequest', '', false, false)]
    local procedure OnSyncRequest(IsInit: Boolean; SyncType: Text[250]; SyncParameters: Text[250])
    var
        MobileNoMapping: Record "ACF Mobile No. Mapping";
    begin
        // ** This function is run before running the initialization and synchronization process.
        // Please note: A synchronization begins after all incoming data from mobile device was processed.
        // The parameter IsInit shows whether it is the first initialization or a synchronization.
        // Available Variables: IsInit, SyncType, SyncParameters
        Mgt.Log('OnSyncRequest', STRSUBSTNO('BEGIN IsInit:%1 SyncType:%2 SyncParameters:%3', IsInit, SyncType, SyncParameters));

        CASE SyncParameters OF
            'confirmReceive':
                BEGIN
                    IF SyncType IN ['SENDONLY', 'SENDONLY_CALCDELTA'] THEN BEGIN
                        IF Configuration.GetBooleanValue('Task Processing Safe Mode') THEN BEGIN
                            Mgt.SendNewNotification(Mgt.GetACFUser(), Text009, FALSE);
                        END ELSE BEGIN
                            Mgt.SendNewNotification(Mgt.GetACFUser(), Text010, FALSE);
                        END;
                    END;
                END;
        END;

        // Checks if Active flag in ACF Mobile No. Mapping is set correctly and resets it if it is not.
        MobileNoMapping.CheckIntegrity(Mgt.GetACFUser);

        Mgt.Log('OnSyncRequest', 'END');
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'OnSetValue', '', false, false)]
    local procedure OnSetValue(sender: Codeunit "ACF App Events"; RecRef: RecordRef; DeviceRecRef: RecordRef; xDeviceRecRef: RecordRef; CalledByInt: Integer; RemoteActionInt: Integer; BCActionInt: Integer; FldRef: FieldRef; DeviceFldRef: FieldRef; xDeviceFldRef: FieldRef; RecordExists: Boolean)
    var
        CalledBy: Option User,Script;
        RemoteAction: Option SetValue,Validate,Modify,ModifyAll;
        BCAction: Option CallTrigger,NoTrigger,Ignore,Reject;
        AvailableFields: Record "Integer" temporary;
        MeetingReport: Record "ACF Meeting Report";
        SalesPresentation: Record "ACF Sales Presentation";
        ServiceItem: Record "Service Item";
        DeviceServiceHeader: Record "Service Header";
        DeviceServiceItemLine: Record "Service Item Line";
        DeviceServiceLine: Record "Service Line";
        DeviceServiceContractLine: Record "Service Contract Line";
        DeviceServiceOrderAllocation: Record "Service Order Allocation";
        ServiceItemComponent: Record "Service Item Component";
        ServiceOrderAllocation: Record "Service Order Allocation";
        InteractionLogEntry: Record "Interaction Log Entry";
        Customer: Record "Customer";
        Vendor: Record "Vendor";
        Contact: Record "Contact";
        SalesHeader: Record "Sales Header";
        DeviceSalesHeader: Record "Sales Header";
        SalesLine: Record "Sales Line";
        DeviceSalesLine: Record "Sales Line";
        PurchaseHeader: Record "Purchase Header";
        DevicePurchaseHeader: Record "Purchase Header";
        PurchaseLine: Record "Purchase Line";
        DevicePurchaseLine: Record "Purchase Line";
        Item: Record "Item";
        DeviceItem: Record "Item";
        ILE: Record "Interaction Log Entry";
        ToDo: Record "To-do";
        Opp: Record "Opportunity";
        ShipToAddress: Record "Ship-to Address";
        CommentLine: Record "Comment Line";
        ServiceContractLine: Record "Service Contract Line";
        DeviceServiceItem: Record "Service Item";
        ACFTimeTrackingLine: Record "ACF Time Tracking Line";
        ACFTimeTrackingJournal: Record "ACF Time Tracking Journal";
        ReleaseSalesDoc: Codeunit "Release Sales Document";
        ReleasePurchaseDoc: Codeunit "Release Purchase Document";
        ItemRecRef: RecordRef;
    begin
        // ** This function is run if the user changes a value in the user interface or Anveo Script uses the commands SETVALUE or VALIDATE.

        // Set Option values        
        CalledBy := CalledByInt;
        RemoteAction := RemoteActionInt;
        BCAction := BCActionInt;

        // Get available Fields
        sender.GetAvailableFields(AvailableFields);

        // If mobile device used VALIDATE in Anveo Script or user entered values in the user interface, run VALIDATE trigger in BC as well.
        IF RemoteAction = RemoteAction::Validate THEN
            BCAction := BCAction::CallTrigger;

        Mgt.Log('OnSetValue', STRSUBSTNO('BEGIN %1.%2 := %3 (Table %4 BCAction:%5 CalledBy:%6)',
          DeviceRecRef.NAME, DeviceFldRef.NAME, FldRefToText(DeviceFldRef), DeviceRecRef.NUMBER, BCAction, CalledBy));

        Mgt.Log('OnSetValue', 'AV:' + format(AvailableFields.Count()));

        // Compare previous value from mobile device with current value in BC to detect data conflicts
        IF RecordExists THEN BEGIN
            // Compare All Fields From Mobile Device With Current Record In BC
            IF NOT CompareMobileAppRecord(xDeviceRecRef, RecRef, AvailableFields) THEN BEGIN
                // BC Record has changed after the device was last synchronized.
                // You can put your conflict management code here.
                Mgt.Log('OnSetValue', 'Current BC data and remote data before changes are different: RecRef <> xDeviceRecRef');
            END;
        END;

        CASE RecRef.NUMBER OF

            Database::Contact:
                BEGIN
                    DeviceRecRef.SetTable(Contact);
                    //Contact No. Matching
                    IF (DeviceFldRef.NUMBER = Contact.FIELDNO("Company No.")) THEN
                        Contact."Company No." := TranslateContactNo(Contact."Company No.");
                    IF (DeviceFldRef.NUMBER = Contact.FIELDNO("No.")) THEN
                        Contact."No." := TranslateContactNo(Contact."No.");
                    DeviceRecRef.GETTABLE(Contact);
                END;

            Database::Customer:
                BEGIN
                    DeviceRecRef.SetTable(Customer);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = Customer.FIELDNO("No.")) THEN
                        Customer."No." := TranslateCustomerNo(Customer."No.");
                    DeviceRecRef.GETTABLE(Customer);
                END;

            Database::Vendor:
                BEGIN
                    DeviceRecRef.SetTable(Vendor);
                    Vendor."No." := TranslateVendorNo(Vendor."No.");
                    DeviceRecRef.GETTABLE(Vendor);
                END;

            // Example: Recalculate FlowFields On Item record
            Database::Item:
                BEGIN
                    DeviceRecRef.SetTable(DeviceItem);
                    IF (DeviceFldRef.NUMBER = DeviceItem.FIELDNO(Inventory)) THEN BEGIN
                        //Recalculate FlowFields For This Item
                        Item.SETRANGE("No.", DeviceItem."No.");
                        ItemRecRef.GETTABLE(Item);
                        Mgt.RECALCFIELDS(ItemRecRef);
                        // Do not change the current value "Inventory"
                        BCAction := BCAction::Ignore;
                    END;
                END;
            // END Example: Recalculate FlowFields on Item record

            Database::"Sales Header":
                BEGIN
                    DeviceRecRef.SetTable(DeviceSalesHeader);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO("Sell-to Customer No.")) THEN
                        DeviceSalesHeader."Sell-to Customer No." := TranslateCustomerNo(DeviceSalesHeader."Sell-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO("Bill-to Customer No.")) THEN
                        DeviceSalesHeader."Bill-to Customer No." := TranslateCustomerNo(DeviceSalesHeader."Bill-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO("Bill-to Contact No.")) THEN
                        DeviceSalesHeader."Sell-to Contact No." := TranslateContactNo(DeviceSalesHeader."Sell-to Contact No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO("Ship-to Code")) THEN
                        DeviceSalesHeader."Ship-to Code" := TranslateShipAddCode(DeviceSalesHeader."Ship-to Code",
                                                                                 DeviceSalesHeader."Sell-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO("Opportunity No.")) THEN
                        DeviceSalesHeader."Opportunity No." := TranslateOppNo(DeviceSalesHeader."Opportunity No.");

                    // Example: Sales Header Release Function
                    IF (DeviceFldRef.NUMBER = DeviceSalesHeader.FIELDNO(Status)) THEN BEGIN
                        CASE DeviceSalesHeader.Status OF
                            DeviceSalesHeader.Status::Released:
                                BEGIN
                                    // Do not set status "Released" directly to Sales Header and send a reject message to reset value on mobile device to
                                    // current BC value.
                                    //BCAction := BCAction::Reject;
                                    // Lookup Matching entry
                                    DeviceRecRef.SetTable(SalesHeader);
                                    TranslateSalesHeader(SalesHeader);
                                    ReleaseSalesDoc.PerformManualRelease(SalesHeader);
                                    RecRef.GETTABLE(SalesHeader);
                                END;

                            DeviceSalesHeader.Status::Open:
                                BEGIN
                                    // Do not set status "Open" directly to Sales Header and send a reject message to reset value on mobile device to
                                    // current BC value.
                                    //Reopen Example See Function: BCFunction SH:REOPEN_REQ or SH:CANCELAPPR_REQ
                                    BCAction := BCAction::Reject;
                                END;
                        END;
                    END;
                    DeviceRecRef.GETTABLE(DeviceSalesHeader);
                END;
            // END Example: Sales Header Release Function

            // Example: Test mobile price calculation of Unit Price in Sales Line with current unit price of BC
            Database::"Sales Line":
                BEGIN
                    DeviceRecRef.SetTable(DeviceSalesLine);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = DeviceSalesLine.FIELDNO("Sell-to Customer No.")) THEN
                        DeviceSalesLine."Sell-to Customer No." := TranslateCustomerNo(DeviceSalesLine."Sell-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesLine.FIELDNO("Bill-to Customer No.")) THEN
                        DeviceSalesLine."Bill-to Customer No." := TranslateCustomerNo(DeviceSalesLine."Bill-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesLine.FIELDNO("Document No.")) THEN
                        DeviceSalesLine."Document No." := TranslateSalesHeaderNo(DeviceSalesLine."Document Type".AsInteger(),
                                                                                 DeviceSalesLine."Document No.");
                    IF (DeviceFldRef.NUMBER = DeviceSalesLine.FIELDNO("Unit Price")) THEN BEGIN
                        RecRef.SetTable(SalesLine);
                        IF SalesLine."Unit Price" <> DeviceSalesLine."Unit Price" THEN BEGIN
                            IF CalledBy = CalledBy::User THEN BEGIN
                                Mgt.Log('OnSetValue', 'Unit Price is different, but OK because it was set by user. This is accepted.');
                            END ELSE BEGIN
                                Mgt.Log('OnSetValue', 'Unit Price is different and was calculated by app price calculation. Manage conflict here.');
                                // Optionally reject local price calculation of mobile device and use price of BC
                                //BCAction := BCAction::Reject;
                            END;
                        END;
                    END;
                    DeviceRecRef.GETTABLE(DeviceSalesLine);
                END;
            // END Example: Test mobile price calculation of Unit Price in Sales Line with current unit price of BC

            Database::"Purchase Header":
                BEGIN
                    DeviceRecRef.SetTable(DevicePurchaseHeader);
                    //Customer No. Matching
                    DevicePurchaseHeader."Pay-to Vendor No." := TranslateVendorNo(DevicePurchaseHeader."Pay-to Vendor No.");
                    DevicePurchaseHeader."Buy-from Vendor No." := TranslateVendorNo(DevicePurchaseHeader."Buy-from Vendor No.");
                    DevicePurchaseHeader."Pay-to Contact No." := TranslateContactNo(DevicePurchaseHeader."Pay-to Contact No.");

                    // Example: Sales Header Release Function
                    IF (DeviceFldRef.NUMBER = DevicePurchaseHeader.FIELDNO(Status)) THEN BEGIN
                        CASE DevicePurchaseHeader.Status OF
                            DevicePurchaseHeader.Status::Released:
                                BEGIN
                                    // Do not set status "Released" directly to Sales Header and send a reject message to reset value on mobile device to
                                    // current BC value.
                                    //BCAction := BCAction::Reject;
                                    // Lookup Matching entry
                                    DeviceRecRef.SetTable(PurchaseHeader);
                                    TranslatePurchaseHeader(PurchaseHeader);
                                    PurchaseHeader.GET(DevicePurchaseHeader."Document Type", DevicePurchaseHeader."No.");
                                    ReleasePurchaseDoc.PerformManualRelease(PurchaseHeader);
                                    RecRef.GETTABLE(PurchaseHeader);
                                END;

                            DevicePurchaseHeader.Status::Open:
                                BEGIN
                                    // Do not set status "Open" directly to Sales Header and send a reject message to reset value on mobile device to
                                    // current BC value.
                                    //Reopen Example See Function: BCFunction SH:REOPEN_REQ or SH:CANCELAPPR_REQ
                                    BCAction := BCAction::Reject;
                                END;
                        END;
                    END;
                    DeviceRecRef.GETTABLE(DevicePurchaseHeader);
                END;

            Database::"Purchase Line":
                BEGIN
                    DeviceRecRef.SetTable(DevicePurchaseLine);
                    //Customer No. Matching
                    DevicePurchaseLine."Buy-from Vendor No." := TranslateVendorNo(DevicePurchaseLine."Buy-from Vendor No.");
                    DevicePurchaseLine."Pay-to Vendor No." := TranslateVendorNo(DevicePurchaseLine."Pay-to Vendor No.");
                    IF (DeviceFldRef.NUMBER = DevicePurchaseLine.FIELDNO("Direct Unit Cost")) THEN BEGIN
                        RecRef.SetTable(PurchaseLine);
                        IF PurchaseLine."Direct Unit Cost" <> DevicePurchaseLine."Direct Unit Cost" THEN BEGIN
                            IF CalledBy = CalledBy::User THEN BEGIN
                                Mgt.Log('OnSetValue', 'Direct Unit Cost is different, but OK because it was set by user. This is accepted.');
                            END ELSE BEGIN
                                Mgt.Log('OnSetValue', 'Direct Unit Cost is different and was calculated by app price calculation. Manage conflict here.');
                                // Optionally reject local price calculation of mobile device and use price of BC
                                //BCAction := BCAction::Reject;
                            END;
                        END;
                    END;
                    DeviceRecRef.GETTABLE(DevicePurchaseLine);
                END;


            Database::"Comment Line":
                BEGIN
                    DeviceRecRef.SetTable(CommentLine);
                    //Customer No. Matching
                    IF CommentLine."Table Name" = CommentLine."Table Name"::Customer THEN BEGIN
                        IF (DeviceFldRef.NUMBER = CommentLine.FIELDNO("No.")) THEN
                            CommentLine."No." := TranslateCustomerNo(CommentLine."No.");
                    END;
                    DeviceRecRef.GETTABLE(CommentLine);
                END;

            Database::"Ship-to Address":
                BEGIN
                    DeviceRecRef.SetTable(ShipToAddress);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = ShipToAddress.FIELDNO("Customer No.")) THEN
                        ShipToAddress."Customer No." := TranslateCustomerNo(ShipToAddress."Customer No.");
                    //Ship-To Adress Code Matching
                    IF (DeviceFldRef.NUMBER = ShipToAddress.FIELDNO(Code)) THEN
                        ShipToAddress.Code := TranslateShipAddCode(ShipToAddress.Code, ShipToAddress."Customer No.");
                    DeviceRecRef.GETTABLE(ShipToAddress);
                END;


            Database::"Interaction Log Entry":
                BEGIN
                    DeviceRecRef.SetTable(ILE);
                    //Contact No. Matching
                    IF (DeviceFldRef.NUMBER = ILE.FIELDNO("Contact No.")) THEN
                        ILE."Contact No." := TranslateContactNo(ILE."Contact No.");
                    IF (DeviceFldRef.NUMBER = ILE.FIELDNO("Contact Company No.")) THEN
                        ILE."Contact Company No." := TranslateContactNo(ILE."Contact Company No.");
                    //Opportunity No. Matching
                    IF (DeviceFldRef.NUMBER = ILE.FIELDNO("Opportunity No.")) THEN
                        ILE."Opportunity No." := TranslateOppNo(ILE."Opportunity No.");
                    DeviceRecRef.GETTABLE(ILE);
                END;

            Database::"To-do":
                BEGIN
                    DeviceRecRef.SetTable(ToDo);
                    //Contact No. Matching
                    IF (DeviceFldRef.NUMBER = ToDo.FIELDNO("Contact No.")) THEN
                        ToDo."Contact No." := TranslateContactNo(ToDo."Contact No.");
                    IF (DeviceFldRef.NUMBER = ToDo.FIELDNO("Contact Company No.")) THEN
                        ToDo."Contact Company No." := TranslateContactNo(ToDo."Contact Company No.");
                    //Opportunity No. Matching
                    IF (DeviceFldRef.NUMBER = ToDo.FIELDNO("Opportunity No.")) THEN
                        ToDo."Opportunity No." := TranslateOppNo(ToDo."Opportunity No.");
                    DeviceRecRef.GETTABLE(ToDo);
                END;

            Database::Opportunity:
                BEGIN
                    DeviceRecRef.SetTable(Opp);
                    //Contact No. Matching
                    IF (DeviceFldRef.NUMBER = Opp.FIELDNO("Contact No.")) THEN
                        Opp."Contact No." := TranslateContactNo(Opp."Contact No.");
                    //Contact Company No. Matching
                    IF (DeviceFldRef.NUMBER = Opp.FIELDNO("Contact Company No.")) THEN
                        Opp."Contact Company No." := TranslateContactNo(Opp."Contact Company No.");
                    IF (DeviceFldRef.NUMBER = Opp.FIELDNO("Sales Document No.")) THEN
                        Opp."Sales Document No." := TranslateSalesHeaderNo(0, Opp."Sales Document No.");
                    Mgt.Log('Opp. Status', FORMAT(Opp.Closed));
                    DeviceRecRef.GETTABLE(Opp);
                END;



            Database::"Service Header":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceHeader);
                    IF (DeviceFldRef.NUMBER = DeviceServiceHeader.FIELDNO("Customer No.")) THEN
                        DeviceServiceHeader."Customer No." := TranslateCustomerNo(DeviceServiceHeader."Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceHeader.FIELDNO("Bill-to Customer No.")) THEN
                        DeviceServiceHeader."Bill-to Customer No." := TranslateCustomerNo(DeviceServiceHeader."Bill-to Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceHeader.FIELDNO("Ship-to Code")) THEN
                        DeviceServiceHeader."Ship-to Code" := TranslateShipAddCode(DeviceServiceHeader."Ship-to Code",
                                                                                 DeviceServiceHeader."Customer No.");

                    DeviceRecRef.GETTABLE(DeviceServiceHeader);
                END;

            Database::"Service Item Line":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceItemLine);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = DeviceServiceItemLine.FIELDNO("Document No.")) THEN
                        DeviceServiceItemLine."Document No." := TranslateServiceHeaderNo(DeviceServiceItemLine."Document Type".AsInteger(),
                                                                                         DeviceServiceItemLine."Document No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceItemLine.FIELDNO("Customer No.")) THEN
                        DeviceServiceItemLine."Customer No." := TranslateCustomerNo(DeviceServiceItemLine."Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceItemLine.FIELDNO("Service Item No.")) THEN
                        DeviceServiceItemLine."Service Item No." := TranslateServiceItemNo(DeviceServiceItemLine."Service Item No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceItemLine.FIELDNO("Ship-to Code")) THEN
                        DeviceServiceItemLine."Ship-to Code" := TranslateShipAddCode(DeviceServiceItemLine."Ship-to Code",
                                                                                   DeviceServiceItemLine."Customer No.");
                    DeviceRecRef.GETTABLE(DeviceServiceItemLine);
                END;

            Database::"Service Line":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceLine);
                    Mgt.Log('AppServiceLine', DeviceServiceLine."No.");
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = DeviceServiceLine.FIELDNO("Document No.")) THEN
                        DeviceServiceLine."Document No." := TranslateServiceHeaderNo(DeviceServiceLine."Document Type".AsInteger(),
                                                                                     DeviceServiceLine."Document No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceLine.FIELDNO("Customer No.")) THEN
                        DeviceServiceLine."Customer No." := TranslateCustomerNo(DeviceServiceItemLine."Customer No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceLine.FIELDNO("Service Item No.")) THEN
                        DeviceServiceLine."Service Item No." := TranslateServiceItemNo(DeviceServiceLine."Service Item No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceLine.FIELDNO("Service Item Line No.")) THEN
                        DeviceServiceLine."Service Item Line No." := TranslateServiceItemLineNo(DeviceServiceLine."Service Item Line No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceLine.FIELDNO("Ship-to Code")) THEN
                        DeviceServiceLine."Ship-to Code" := TranslateShipAddCode(DeviceServiceLine."Ship-to Code",
                                                                                 DeviceServiceLine."Customer No.");


                    DeviceRecRef.GETTABLE(DeviceServiceLine);
                END;

            Database::"Service Contract Line":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceContractLine);
                    Mgt.Log('DeviceServiceCL', DeviceServiceContractLine."Service Item No.");
                    IF (DeviceFldRef.NUMBER = DeviceServiceContractLine.FIELDNO("Service Item No.")) THEN
                        DeviceServiceContractLine."Service Item No." := TranslateServiceItemNo(DeviceServiceContractLine."Service Item No.");
                    Mgt.Log('DeviceServiceCL', DeviceServiceContractLine."Service Item No.");

                    IF (DeviceFldRef.NUMBER = DeviceServiceContractLine.FIELDNO("Service Item No.")) THEN BEGIN
                        RecRef.SetTable(DeviceServiceContractLine);
                        IF BCAction IN [BCAction::CallTrigger, BCAction::NoTrigger] THEN BEGIN
                            //DeviceServiceContractLine."Service Item No." := DeviceFldRef.VALUE;
                            IF BCAction = BCAction::CallTrigger THEN BEGIN
                                DeviceServiceContractLine.HideDialogBox(TRUE);
                                DeviceServiceContractLine."Service Item No." := TranslateServiceItemNo(FORMAT(DeviceFldRef.VALUE));
                                Mgt.Log('DeviceServiceCL', DeviceServiceContractLine."Service Item No.");
                                DeviceServiceContractLine.VALIDATE("Service Item No.");
                            END;
                            RecRef.GETTABLE(DeviceServiceContractLine);
                        END;
                        BCAction := BCAction::Reject;
                    END;
                    DeviceRecRef.GETTABLE(DeviceServiceContractLine);
                END;


            Database::"Service Item":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceItem);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = DeviceServiceItem.FIELDNO("Customer No.")) THEN
                        DeviceServiceItem."Customer No." := TranslateCustomerNo(DeviceServiceItem."Customer No.");
                    IF DeviceFldRef.NUMBER = DeviceServiceItem.FIELDNO("Item No.") THEN BEGIN
                        BCAction := BCAction::Reject;
                        ServiceItem.GET(TranslateServiceItemNo(DeviceServiceItem."No."));
                        ServiceItem.OmitAssignResSkills(TRUE);
                        ServiceItem.VALIDATE("Item No.", DeviceServiceItem."Item No.");
                        RecRef.GETTABLE(DeviceServiceItem);
                    END;
                    DeviceRecRef.GETTABLE(DeviceServiceItem);
                END;

            Database::"Service Order Allocation":
                BEGIN
                    DeviceRecRef.SetTable(DeviceServiceOrderAllocation);
                    IF (DeviceFldRef.NUMBER = DeviceServiceOrderAllocation.FIELDNO("Resource No.")) THEN BEGIN
                        BCAction := BCAction::Reject;
                        DeviceRecRef.SetTable(ServiceOrderAllocation);
                        ServiceOrderAllocation.GET(DeviceServiceOrderAllocation."Entry No.");
                        ServiceOrderAllocation.SetHideDialog(TRUE);
                        ServiceOrderAllocation.VALIDATE("Resource No.", DeviceServiceOrderAllocation."Resource No.");
                        RecRef.GETTABLE(ServiceOrderAllocation);
                    END;
                END;

            Database::"Service Item Component":
                BEGIN
                    DeviceRecRef.SetTable(ServiceItemComponent);
                    IF (DeviceFldRef.NUMBER = ServiceItemComponent.FIELDNO("Parent Service Item No.")) THEN
                        ServiceItemComponent."Parent Service Item No." := TranslateServiceItemNo(ServiceItemComponent."Parent Service Item No.");
                    IF (DeviceFldRef.NUMBER = ServiceItemComponent.FIELDNO("No.")) THEN
                        ServiceItemComponent."No." := TranslateServiceItemNo(ServiceItemComponent."No.");
                    DeviceRecRef.GETTABLE(ServiceItemComponent);
                END;

            Database::"ACF Sales Presentation":
                BEGIN
                    DeviceRecRef.SetTable(SalesPresentation);
                    //Customer No. Matching
                    IF (DeviceFldRef.NUMBER = SalesPresentation.FIELDNO("Document No.")) THEN
                        SalesPresentation."Document No." := TranslateSalesHeaderNo(SalesPresentation."Document Type",
                                                                                   SalesPresentation."Document No.");
                    DeviceRecRef.GETTABLE(SalesPresentation);
                END;

            //To Use ACF Meeting Reports please Import Anveo APP Sales FOB
            Database::"ACF Meeting Report":
                BEGIN
                    DeviceRecRef.SetTable(MeetingReport);
                    //Contact No. Matching
                    IF (DeviceFldRef.NUMBER = MeetingReport.FIELDNO("Contact No.")) THEN
                        MeetingReport."Contact No." := TranslateContactNo(MeetingReport."Contact No.");
                    IF (DeviceFldRef.NUMBER = MeetingReport.FIELDNO("Contact Company No.")) THEN
                        MeetingReport."Contact Company No." := TranslateContactNo(MeetingReport."Contact Company No.");
                    IF (DeviceFldRef.NUMBER = MeetingReport.FIELDNO("Opportunity No.")) THEN
                        MeetingReport."Opportunity No." := TranslateOppNo(MeetingReport."Opportunity No.");
                    DeviceRecRef.GETTABLE(MeetingReport);
                END;

            Database::"ACF Time Tracking Line":
                BEGIN
                    DeviceRecRef.SetTable(ACFTimeTrackingLine);
                    ACFTimeTrackingLine."Document No." := TranslateServiceHeaderNo(ACFTimeTrackingLine."Document Type",
                                                                                   ACFTimeTrackingLine."Document No.");
                    DeviceRecRef.GETTABLE(ACFTimeTrackingLine);
                END;

            Database::"ACF Time Tracking Journal":
                BEGIN
                    DeviceRecRef.SetTable(ACFTimeTrackingJournal);
                    ACFTimeTrackingJournal."Active Document No." := TranslateServiceHeaderNo(1, ACFTimeTrackingJournal."Active Document No.");
                    DeviceRecRef.GETTABLE(ACFTimeTrackingJournal);
                END;

        // Insert Your Code To Handle Incoming Data Here
        // ...
        END;

        IF BCAction IN [BCAction::CallTrigger, BCAction::NoTrigger] THEN BEGIN
            // *** SET VALUE FROM MOBILE DEVICE TO BC RECORD ***
            FldRef.VALUE := DeviceFldRef.VALUE;
            // Run VALIDATE Trigger In BC?
            IF BCAction = BCAction::CallTrigger THEN BEGIN
                FldRef.VALIDATE;
            END;
        END;
        // Current Field was processed. For further xDeviceRecRef testing use current value.
        xDeviceFldRef.VALUE := FldRef.VALUE;
        IF RecRef.NUMBER = Database::"Service Order Allocation" THEN BEGIN
            RecRef.SetTable(ServiceOrderAllocation);
            Mgt.Log('OnSetValue', FORMAT(ServiceOrderAllocation.Status));
        END;
        Mgt.Log('OnSetValue', STRSUBSTNO('END BCAction:%1', BCAction));
    end;

    local procedure OnBeforeInsert(var RecRef: RecordRef)
    var
        MeetingReport: Record "ACF Meeting Report";
        SalesPresentation: Record "ACF Sales Presentation";
        SalesPresentationLine: Record "ACF Sales Presentation Line";
        ServiceLine: Record "Service Line";
        ServiceItem: Record "Service Item";
        ServiceItemComponent: Record "Service Item Component";
        ServiceOrderAllocation: Record "Service Order Allocation";
        ServiceContractLine: Record "Service Contract Line";
        Opp: Record "Opportunity";
        OpportunityEntry: Record "Opportunity Entry";
        ILE: Record "Interaction Log Entry";
        BC_ILE: Record "Interaction Log Entry";
        Contact: Record "Contact";
        CIG: Record "Contact Industry Group";
        CMG: Record "Contact Mailing Group";
        RMCL: Record "Rlshp. Mgt. Comment Line";
        SalesHeader: Record "Sales Header";
        SalesLine: Record "Sales Line";
        PurchaseHeader: Record "Purchase Header";
        PurchaseLine: Record "Purchase Line";
        PurchaseCommentLine: Record "Purch. Comment Line";
        SalesCommentLine: Record "Sales Comment Line";
        ServiceHeader: Record "Service Header";
        ContactProfileAnswer: Record "Contact Profile Answer";
        FileRep: Record "ACF File Repository";
        FileRep2: Record "ACF File Repository";
        Customer: Record "Customer";
        Vendor: Record "Vendor";
        ToDo: Record "To-do";
        ShipToAddress: Record "Ship-to Address";
        ShipToAddress2: Record "Ship-to Address";
        CommentLine: Record "Comment Line";
        ServiceCommentLine: Record "Service Comment Line";
        ServiceItemLine: Record "Service Item Line";
        ACFGPS: Record "ACF GPS";
        ACFGPS2: Record "ACF GPS";
        LocalRecRef: RecordRef;
        TranslateRecRef: RecordRef;
        entryNo: Integer;
    begin

        // This function is run before RecRef will be inserted into BC database.
        // It is called by OnInsert.
        Mgt.Log('OnBeforeInsertRec', STRSUBSTNO('BEGIN Table %1 %2', RecRef.NUMBER, RecRef.NAME));
        // Mobile No. Series Management
        CASE RecRef.NUMBER OF

            Database::Customer:
                BEGIN
                    RecRef.SetTable(Customer);
                    Customer."No." := '';
                    RecRef.GETTABLE(Customer);
                END;

            Database::Contact:
                BEGIN
                    RecRef.SetTable(Contact);
                    //Remove Temporary Number To Use BC Number Series Management
                    Contact."Company No." := TranslateContactNo(Contact."Company No.");
                    Contact."No." := '';
                    RecRef.GETTABLE(Contact);
                END;

            Database::"Contact Industry Group":
                BEGIN
                    RecRef.SetTable(CIG);
                    CIG."Contact No." := TranslateContactNo(CIG."Contact No.");
                    RecRef.GETTABLE(CIG);
                END;

            Database::"Contact Mailing Group":
                BEGIN
                    RecRef.SetTable(CMG);
                    CMG."Contact No." := TranslateContactNo(CMG."Contact No.");
                    RecRef.GETTABLE(CMG);
                END;

            Database::Vendor:
                BEGIN
                    RecRef.SetTable(Vendor);
                    Vendor."No." := '';
                    RecRef.GETTABLE(Vendor);
                END;

            Database::"Rlshp. Mgt. Comment Line":
                BEGIN
                    RecRef.SetTable(RMCL);
                    //Contact No. Matching
                    IF RMCL."Table Name" = RMCL."Table Name"::Contact THEN
                        RMCL."No." := TranslateContactNo(RMCL."No.");
                    //Opportunity No. Matching
                    IF RMCL."Table Name" = RMCL."Table Name"::Opportunity THEN
                        RMCL."No." := TranslateOppNo(RMCL."No.");
                    RecRef.GETTABLE(RMCL);
                END;

            Database::"Ship-to Address":
                BEGIN
                    RecRef.SetTable(ShipToAddress);
                    ShipToAddress."Customer No." := TranslateCustomerNo(ShipToAddress."Customer No.");
                    ShipToAddress2.RESET;
                    ShipToAddress2.SETRANGE("Customer No.", ShipToAddress."Customer No.");
                    ShipToAddress2.SETFILTER(Code, 'SHIP????');
                    IF ShipToAddress2.FINDLAST THEN
                        ShipToAddress.Code := INCSTR(ShipToAddress2.Code)
                    ELSE
                        ShipToAddress.Code := 'SHIP0001';
                    RecRef.GETTABLE(ShipToAddress);
                END;

            Database::"Comment Line":
                BEGIN
                    RecRef.SetTable(CommentLine);
                    IF CommentLine."Table Name" = CommentLine."Table Name"::Customer THEN
                        CommentLine."No." := TranslateCustomerNo(CommentLine."No.");
                    RecRef.GETTABLE(CommentLine);
                END;

            //To Use ACF Meeting Reports please Import Anveo APP Sales FOB
            Database::"ACF Meeting Report":
                BEGIN
                    RecRef.SetTable(MeetingReport);
                    //Contact No. Matching
                    MeetingReport."Entry No." := 0;
                    MeetingReport."Contact No." := TranslateContactNo(MeetingReport."Contact No.");
                    MeetingReport."Contact Company No." := TranslateContactNo(MeetingReport."Contact Company No.");
                    MeetingReport."Opportunity No." := TranslateOppNo(MeetingReport."Opportunity No.");
                    RecRef.GETTABLE(MeetingReport);
                END;

            Database::"Interaction Log Entry":
                BEGIN
                    //Find Next EntryNo
                    BC_ILE.LOCKTABLE;
                    BC_ILE.FINDLAST();
                    RecRef.SetTable(ILE);
                    ILE."Entry No." := BC_ILE."Entry No." + 1;
                    //Contact No. Matching
                    ILE."Contact No." := TranslateContactNo(ILE."Contact No.");
                    ILE."Contact Company No." := TranslateContactNo(ILE."Contact Company No.");
                    //Opportunity No. Matching
                    ILE."Opportunity No." := TranslateOppNo(ILE."Opportunity No.");
                    RecRef.GETTABLE(ILE);
                END;

            Database::"To-do":
                BEGIN
                    RecRef.SetTable(ToDo);
                    //Contact No. Matching
                    ToDo."Contact No." := TranslateContactNo(ToDo."Contact No.");
                    ToDo."Contact Company No." := TranslateContactNo(ToDo."Contact Company No.");
                    //Opportunity No. Matching
                    ToDo."Opportunity No." := TranslateOppNo(ToDo."Opportunity No.");
                    //Remove Temporary Number To Use BC Number Series Management
                    ToDo."No." := '';
                    RecRef.GETTABLE(ToDo);
                END;

            Database::Opportunity:
                BEGIN
                    RecRef.SetTable(Opp);
                    //Contact No. Matching
                    Opp."Contact No." := TranslateContactNo(Opp."Contact No.");
                    //Opportunity No. Matching
                    Opp."Contact Company No." := TranslateContactNo(Opp."Contact Company No.");
                    //Remove Temporary Number To Use BC Number Series Management
                    Opp."No." := '';
                    RecRef.GETTABLE(Opp);
                END;

            Database::"Opportunity Entry":
                BEGIN
                    entryNo := 1;
                    IF OpportunityEntry.FINDLAST THEN
                        entryNo := entryNo + OpportunityEntry."Entry No.";
                    RecRef.SetTable(OpportunityEntry);
                    OpportunityEntry."Contact No." := TranslateContactNo(OpportunityEntry."Contact No.");
                    OpportunityEntry."Contact Company No." := TranslateContactNo(OpportunityEntry."Contact Company No.");
                    OpportunityEntry."Opportunity No." := TranslateOppNo(OpportunityEntry."Opportunity No.");
                    OpportunityEntry."Entry No." := entryNo;
                    RecRef.GETTABLE(OpportunityEntry);
                END;


            // Remove Mobile No. Series to activate automatic no. series management from BC
            Database::"Sales Header":
                BEGIN
                    RecRef.SetTable(SalesHeader);
                    //Customer No. Matching
                    SalesHeader."Sell-to Customer No." := TranslateCustomerNo(SalesHeader."Sell-to Customer No.");
                    SalesHeader."Bill-to Customer No." := TranslateCustomerNo(SalesHeader."Bill-to Customer No.");
                    //Ship-To Adress Code Matching
                    SalesHeader."Ship-to Code" := TranslateShipAddCode(SalesHeader."Ship-to Code", SalesHeader."Sell-to Customer No.");
                    //Remove Temporary Number To Use BC Number Series Management
                    SalesHeader."No." := '';
                    RecRef.GETTABLE(SalesHeader);
                END;
            // Use already existing mapping for Sales Header also for Sales Lines
            Database::"Sales Line":
                BEGIN
                    RecRef.SetTable(SalesLine);
                    //Sales Header Document Type and Document No. Matching
                    SalesLine."Document No." := TranslateSalesHeaderNo(SalesLine."Document Type".AsInteger(), SalesLine."Document No.");
                    //Customer No. Matching
                    SalesLine."Sell-to Customer No." := TranslateCustomerNo(SalesLine."Sell-to Customer No.");
                    SalesLine."Bill-to Customer No." := TranslateCustomerNo(SalesLine."Bill-to Customer No.");
                    RecRef.GETTABLE(SalesLine);
                END;

            Database::"Purchase Header":
                BEGIN
                    RecRef.SetTable(PurchaseHeader);
                    //Vendor No. Matching
                    PurchaseHeader."Buy-from Vendor No." := TranslateVendorNo(PurchaseHeader."Buy-from Vendor No.");
                    PurchaseHeader."Pay-to Vendor No." := TranslateVendorNo(PurchaseHeader."Pay-to Vendor No.");
                    PurchaseHeader."No." := '';
                    RecRef.GETTABLE(PurchaseHeader);
                END;

            Database::"Purchase Line":
                BEGIN
                    RecRef.SetTable(PurchaseLine);
                    PurchaseLine."Document No." := TranslatePurchaseHeaderNo(PurchaseLine."Document Type".AsInteger(), PurchaseLine."Document No.");
                    PurchaseLine."Buy-from Vendor No." := TranslateVendorNo(PurchaseLine."Buy-from Vendor No.");
                    PurchaseLine."Pay-to Vendor No." := TranslateVendorNo(PurchaseLine."Pay-to Vendor No.");
                    RecRef.GETTABLE(PurchaseLine);
                END;

            Database::"Purch. Comment Line":
                BEGIN
                    RecRef.SetTable(PurchaseCommentLine);
                    PurchaseCommentLine."No." := TranslatePurchaseHeaderNo(PurchaseCommentLine."Document Type".AsInteger(), PurchaseCommentLine."No.");
                    RecRef.GETTABLE(PurchaseCommentLine);
                END;

            Database::"Sales Comment Line":
                BEGIN
                    RecRef.SetTable(SalesCommentLine);
                    SalesCommentLine."No." := TranslateSalesHeaderNo(SalesCommentLine."Document Type".AsInteger(), SalesCommentLine."No.");
                    RecRef.GETTABLE(SalesCommentLine);
                END;

            // Remove Mobile No. Series to activate automatic no. series management from BC
            Database::"Service Header":
                BEGIN
                    RecRef.SetTable(ServiceHeader);
                    ServiceHeader."Document Date" := TODAY;
                    //Customer No. Matching
                    ServiceHeader."Customer No." := TranslateCustomerNo(ServiceHeader."Customer No.");
                    ServiceHeader."Bill-to Customer No." := TranslateCustomerNo(ServiceHeader."Bill-to Customer No.");
                    //Ship-To Adress Code Matching
                    ServiceHeader."Ship-to Code" := TranslateShipAddCode(ServiceHeader."Ship-to Code", ServiceHeader."Customer No.");
                    //Remove Temporary Number To Use BC Number Series Management
                    ServiceHeader."No." := '';
                    RecRef.GETTABLE(ServiceHeader);
                END;

            Database::"Service Item Line":
                BEGIN
                    RecRef.SetTable(ServiceItemLine);
                    ServiceItemLine."Document No." := TranslateServiceHeaderNo(ServiceItemLine."Document Type".AsInteger(), ServiceItemLine."Document No.");
                    ServiceItemLine."Customer No." := TranslateCustomerNo(ServiceItemLine."Customer No.");
                    ServiceItemLine."Service Item No." := TranslateServiceItemNo(ServiceItemLine."Service Item No.");
                    RecRef.GETTABLE(ServiceItemLine);
                END;

            Database::"Service Line":
                BEGIN
                    RecRef.SetTable(ServiceLine);
                    ServiceLine."Document No." := TranslateServiceHeaderNo(ServiceLine."Document Type".AsInteger(), ServiceLine."Document No.");
                    ServiceLine."Customer No." := TranslateCustomerNo(ServiceLine."Customer No.");
                    ServiceLine."No." := TranslateServiceItemNo(ServiceLine."No.");
                    ServiceLine."Service Item Line No." := TranslateServiceItemLineNo(ServiceLine."Service Item Line No.");
                    RecRef.GETTABLE(ServiceLine);
                END;


            Database::"ACF File Repository":
                BEGIN
                    // Get Next EntryNo To Avoid Conflicts
                    RecRef.SetTable(FileRep);
                    FileRep2.LOCKTABLE;
                    FileRep2.SETRANGE("Table No.", FileRep."Table No.");
                    FileRep2.SETRANGE("Document Type", FileRep."Document Type");
                    FileRep2.SETRANGE(Code, FileRep.Code);
                    FileRep2.SETRANGE("Line No.", FileRep."Line No.");
                    IF FileRep2.FINDLAST THEN
                        FileRep."Entry No." := FileRep2."Entry No." + 1;
                    //Data Matching For Customer Table
                    IF FileRep."Table No." = Database::Customer THEN
                        FileRep.Code := TranslateCustomerNo(FileRep.Code);
                    //Data Matching For Sales Header Table
                    IF FileRep."Table No." = Database::Contact THEN
                        FileRep.Code := TranslateContactNo(FileRep.Code);
                    //Data Matching For Sales Header Table
                    IF FileRep."Table No." = Database::"Sales Header" THEN BEGIN
                        //Translate Sales Header No. From Table Sales Header
                        FileRep.Code := TranslateSalesHeaderNo(FileRep."Document Type", FileRep.Code);
                    END;
                    IF FileRep."Table No." = Database::"Service Header" THEN BEGIN
                        //Translate Service Header No. From Table Service Header
                        FileRep.Code := TranslateServiceHeaderNo(FileRep."Document Type", FileRep.Code);
                    END;
                    IF FileRep."Table No." = Database::"Service Item Line" THEN BEGIN
                        FileRep.Code := TranslateServiceHeaderNo(FileRep."Document Type", FileRep.Code);
                    END;
                    IF FileRep."Table No." = Database::"ACF Meeting Report" THEN BEGIN
                        //Translate Meeting Report Entry No. and Contact No.
                        FileRep.Code := TranslateContactNo(FileRep.Code);
                        FileRep."Line No." := TranslateMeetingReportNo(FileRep."Line No.");
                    END;

                    RecRef.GETTABLE(FileRep);
                END;

            Database::"Service Comment Line":
                BEGIN
                    RecRef.SetTable(ServiceCommentLine);
                    IF ServiceCommentLine."Table Name" = ServiceCommentLine."Table Name"::"Service Header" THEN
                        ServiceCommentLine."No." := TranslateServiceHeaderNo(ServiceCommentLine."Table Subtype", ServiceCommentLine."No.");
                    RecRef.GETTABLE(ServiceCommentLine);
                END;

            Database::"Service Item":
                BEGIN
                    RecRef.SetTable(ServiceItem);
                    ServiceItem."No." := '';
                    RecRef.GETTABLE(ServiceItem);
                END;

            Database::"Service Item Component":
                BEGIN
                    RecRef.SetTable(ServiceItemComponent);
                    ServiceItemComponent."Parent Service Item No." := TranslateServiceItemNo(ServiceItemComponent."Parent Service Item No.");
                    ServiceItemComponent."No." := TranslateServiceItemNo(ServiceItemComponent."No.");
                    RecRef.GETTABLE(ServiceItemComponent);
                END;

            Database::"Service Order Allocation":
                BEGIN
                    RecRef.SetTable(ServiceOrderAllocation);
                    ServiceOrderAllocation."Document No." := TranslateServiceHeaderNo(ServiceOrderAllocation."Document Type".AsInteger(),
                      ServiceOrderAllocation."Document No.");
                    ServiceOrderAllocation."Service Item No." := TranslateServiceItemNo(ServiceOrderAllocation."Service Item No.");
                    RecRef.GETTABLE(ServiceOrderAllocation);
                END;

            Database::"Service Contract Line":
                BEGIN
                    RecRef.SetTable(ServiceContractLine);
                    ServiceContractLine."Service Item No." := TranslateServiceItemNo(ServiceContractLine."Service Item No.");
                    RecRef.GETTABLE(ServiceContractLine);
                END;

            Database::"Contact Profile Answer":
                BEGIN
                    RecRef.SetTable(ContactProfileAnswer);
                    ContactProfileAnswer."Contact No." := TranslateContactNo(ContactProfileAnswer."Contact No.");
                    ContactProfileAnswer."Contact Company No." := TranslateContactNo(ContactProfileAnswer."Contact Company No.");
                    RecRef.GETTABLE(ContactProfileAnswer);
                END;

            Database::"ACF Sales Presentation":
                BEGIN
                    RecRef.SetTable(SalesPresentation);
                    SalesPresentation."Customer No." := TranslateCustomerNo(SalesPresentation."Customer No.");
                    RecRef.GETTABLE(SalesPresentation);
                END;

            Database::"ACF Sales Presentation Line":
                BEGIN
                    RecRef.SetTable(SalesPresentationLine);
                    SalesPresentationLine."Customer No." := TranslateCustomerNo(SalesPresentationLine."Customer No.");
                    RecRef.GETTABLE(SalesPresentationLine);
                END;

            Database::"ACF GPS":
                BEGIN
                    RecRef.SetTable(ACFGPS);
                    ACFGPS2.SETRANGE("Anveo User Device ID", ACFGPS."Anveo User Device ID");
                    ACFGPS2.SETRANGE("Keep on Device", TRUE);
                    ACFGPS2.MODIFYALL("Keep on Device", FALSE);
                    ACFGPS.Received := CURRENTDATETIME;
                    ACFGPS."Keep on Device" := TRUE;
                    RecRef.GETTABLE(ACFGPS);
                END;



        // Insert Your Mobile No. Series Mappings For Other Tables Here
        // ...

        END;
        Mgt.Log('OnBeforeInsertRec', STRSUBSTNO('END'));
        // END Mobile No. Series Management        
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'OnInsertRec', '', false, false)]
    local procedure OnInsertRec(sender: Codeunit "ACF App Events"; RecRef: RecordRef; DeviceRecRef: RecordRef; CalledByInt: Integer; BCActionInt: Integer; RemoteActionInt: Integer)
    var
        FileRepository: Record "ACF File Repository";
        AvailableFields: Record "Integer" temporary;
        CalledBy: Option User,Script;
        RemoteAction: Option SetValue,Validate,Modify,ModifyAll;
        BCAction: Option CallTrigger,NoTrigger,Ignore,Reject;
        InsertSuccessful: Boolean;
    begin
        // ** This function is run if Anveo Script uses the command Record:INSERT()
        // Available Variables: RecRef, DeviceRecRef, CalledBy, BCAction

        // Set Option values        
        CalledBy := CalledByInt;
        RemoteAction := RemoteActionInt;
        BCAction := BCActionInt;

        // Get available Fields
        sender.GetAvailableFields(AvailableFields);

        Mgt.Log('OnInsertRec', STRSUBSTNO('BEGIN Table %1 %2 BCAction:%3', RecRef.NUMBER, RecRef.NAME, BCAction));

        OnBeforeInsert(RecRef);

        // Insert Your Code Here...
        // ...
        // ** Example Reject Inserts on mobile device under special conditions
        // IF Condition THEN
        //   BCAction := BCAction::Reject;

        IF BCAction IN [BCAction::CallTrigger, BCAction::NoTrigger] THEN BEGIN
            // *** INSERT RECORD TO BC Database ***
            InsertSuccessful := RecRef.INSERT(BCAction = BCAction::CallTrigger);

            CASE RecRef.NUMBER OF
                // Print Report Example
                Database::"ACF File Repository":
                    BEGIN
                        RecRef.SetTable(FileRepository);
                        //Using Type For Different Tasks
                        CASE FileRepository.Type OF
                            //Type: "Report Request" For Report Demand From App
                            FileRepository.Type::"Report Request":
                                PrintReport(FileRepository);
                        END;
                    END;
            END;

            IF NOT InsertSuccessful THEN BEGIN
                // Handle errors
                CASE RecRef.NUMBER OF
                    Database::"Sales Header":
                        BEGIN
                            ERROR(Error001, RecRef.CAPTION, RecRef.RECORDID);
                        END;
                END;
            END;
        END;
        Mgt.Log('OnInsertRec', STRSUBSTNO('END BCAction:%1 InsertSuccessful:%2', BCAction, InsertSuccessful));
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'OnModifyRec', '', false, false)]
    local procedure OnModifyRec(sender: Codeunit "ACF App Events"; RecRef: RecordRef; DeviceRecRef: RecordRef; xDeviceRecRef: RecordRef; FldRef: FieldRef; DeviceFldRef: FieldRef; xDeviceFldRef: FieldRef; CalledByInt: Integer; BCActionInt: Integer; RemoteActionInt: Integer)
    var

        CalledBy: Option User,Script;
        RemoteAction: Option SetValue,Validate,Modify,ModifyAll;
        BCAction: Option CallTrigger,NoTrigger,Ignore,Reject;
        ModifySuccessful: Boolean;
        AvailableFields: Record "Integer" temporary;

        Customer: Record Customer;
        Contact: Record Contact;
        Vendor: Record Vendor;
        UpdateContFromCust: Codeunit "CustCont-Update";
        UpdateCustVendBank: Codeunit "CustVendBank-Update";
        UpdateContFromVendor: Codeunit "VendCont-Update";

    begin
        // ** This function is run if the user has changed a value in the user interface or if Anveo Script used the command
        //    Record:MODIFY()
        // Available Variables: RecRef, DeviceRecRef, xDeviceRecRef, BCAction, CalledBy
        // In case of RemoteAction = ModifyAll: FldRef, DeviceFldRef, xDeviceFldRef

        // Set Option values        
        CalledBy := CalledByInt;
        BCAction := BCActionInt;
        RemoteAction := RemoteActionInt;

        // Get available Fields
        sender.GetAvailableFields(AvailableFields);

        Mgt.Log('OnModifyRec', STRSUBSTNO('BEGIN Table %1 %2 BCAction:%3 CalledBy:%4',
          RecRef.NUMBER, RecRef.NAME, BCAction, CalledBy));

        // If RemoteAction is ModifyAll then set the changed value from the device record to the BC record
        Mgt.Log('OnModifyRec', 'Remoteaction: ' + FORMAT(RemoteAction));
        IF RemoteAction = RemoteAction::ModifyAll THEN BEGIN
            FldRef.VALUE := DeviceFldRef.VALUE;
        END;

        // Insert Your Code Here...
        // ...
        // ** Example Reject Changes on mobile device under special conditions
        // IF Condition THEN
        //   BCAction := BCAction::Reject;

        IF BCAction IN [BCAction::CallTrigger, BCAction::NoTrigger] THEN BEGIN
            // *** MODIFY BC RECORD ***
            ModifySuccessful := RecRef.MODIFY(BCAction = BCAction::CallTrigger);

            IF ModifySuccessful THEN BEGIN
                CASE RecRef.NUMBER OF
                    Database::Customer:
                        BEGIN
                            RecRef.SetTable(Customer);
                            UpdateContFromCust.OnModify(Customer);
                        END;
                    Database::Vendor:
                        BEGIN
                            RecRef.SetTable(Vendor);
                            UpdateContFromVendor.OnModify(Vendor);
                        END;

                    Database::Contact:
                        BEGIN
                            RecRef.SetTable(Contact);
                            UpdateCustVendBank.RUN(Contact);
                        END;
                END;
            END;

            IF NOT ModifySuccessful THEN BEGIN
                // Handle errors

            END;
        END;

        Mgt.Log('OnModifyRec', STRSUBSTNO('END BCAction:%1 ModifySuccessful:%2', BCAction, ModifySuccessful));
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'OnDeleteRec', '', false, false)]
    local procedure OnDeleteRec(sender: Codeunit "ACF App Events"; RecRef: RecordRef; DeviceRecRef: RecordRef; CalledByInt: Integer; BCActionInt: Integer; RemoteActionInt: Integer)
    var

        CalledBy: Option User,Script;
        BCAction: Option CallTrigger,NoTrigger,Ignore,Reject;
        RemoteAction: Option SetValue,Validate,Modify,ModifyAll;
        DeleteSuccessful: Boolean;
        AvailableFields: Record "Integer" temporary;
    begin
        // ** This function is run if Anveo Script uses the command Record:DELETE().
        // Available Variables: RecRef, DeviceRecRef, BCAction, CalledBy

        // Set Option values        
        CalledBy := CalledByInt;
        RemoteAction := RemoteActionInt;
        BCAction := BCActionInt;

        // Get available Fields
        sender.GetAvailableFields(AvailableFields);

        Mgt.Log('OnDeleteRec', STRSUBSTNO('BEGIN Table %1 %2 BCAction:%3 CalledBy:%4',
          RecRef.NUMBER, RecRef.NAME, BCAction, CalledBy));


        // Insert Your Code Here...
        // ...
        // ** Example Reject Deletions on mobile device under special conditions
        // IF Condition THEN
        //   BCAction := BCAction::Reject;

        IF BCAction IN [BCAction::CallTrigger, BCAction::NoTrigger] THEN BEGIN
            // *** DELETE RECORD FROM BC Database ***
            DeleteSuccessful := RecRef.DELETE(BCAction = BCAction::CallTrigger);

            IF NOT DeleteSuccessful THEN BEGIN
                // Handle errors
            END;
        END;
        Mgt.Log('OnDeleteRec', STRSUBSTNO('END BCAction:%1 DeleteSuccessful:%2', BCAction, DeleteSuccessful));
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'CalcOnlineFieldValue', '', false, false)]
    local procedure CalcOnlineFieldValue(RecRef: RecordRef; var FldRef: FieldRef; AppField: Record "ACF App Field"; var ResultValue: Variant)
    var
        Item: Record Item;
    begin

        // ** For virtual fields set ResultValue to the desired value
        // For regular BC fields, set FldRef.VALUE

        CASE RecRef.NUMBER OF
            27:
                BEGIN
                    CASE AppField."Field Name" OF
                        'Inventory BLUE':
                            BEGIN
                                RecRef.SetTable(Item);
                                IF Mgt.GetACFUser() = 'SALES_DE' THEN
                                    Item.SETRANGE("Location Filter", 'BLAU')
                                ELSE
                                    IF Mgt.GetACFUser() = 'SALES_EN' THEN
                                        Item.SETRANGE("Location Filter", 'BLUE');
                                Item.CALCFIELDS(Inventory);
                                ResultValue := Item.Inventory;
                            END;
                    END;
                    //CASE FldRef.NUMBER OF
                    //  Item.FIELDNO(Description): FldRef.VALUE := 'Changed Description';
                    //END;
                END;
        END;
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'LanguageIDMapping', '', false, false)]
    local procedure LanguageIDMapping(var LanguageID: Integer)
    begin
        // For some languages it is required for the mobile device to transform the windows language of table Language into another
        // windows language value.
        //case LanguageID of
        //    1034:
        //        LanguageID := 3082;
        //end;
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'GetSyncPackageUserFilter', '', false, false)]
    local procedure GetSyncPackageUserFilter(AnveoUserCode: Code[20]; FilterCode: Code[20]; var UserFilter: Text[1024])
    begin
        // Syntax: %CUSTOMFILTER:MYFILTERCODE in Synchronization Package returns a filter string via App Events.
        // MYFILTERCODE is the identifier of the filter value.
        // Please note: this filter will be updated on a manual Update Delta Scripts only, not every time a user
        // synchronizes.
        Mgt.Log('GetFilter', AnveoUserCode + ' ' + FilterCode);
        if (AnveoUserCode = 'SALES_EN') AND (FilterCode = 'MYTESTFILTER') then UserFilter := 'BLABLA';
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'SetFlowfilter', '', false, false)]
    local procedure SetFlowfilter(var RecRef: RecordRef)
    begin
        // Use this function to set flowfilters to flowfields which are being calculated during initialization/synchronization
        // see examples below

        CASE RecRef.NUMBER OF

        //  Resource based Filter on Items in Service
        //  Database::Item: BEGIN
        //    ACFUser.GET(Mgt.GetACFUser());
        //    IF ACFUser."Relation Type" = ACFUser."Relation Type"::Resource THEN BEGIN
        //      ResourceLocation.SETRANGE(ResourceLocation."Resource No.",ACFUser."Relation Code");
        //      ResourceLocation.SETFILTER("Starting Date", '%1|<=%2', 0D, TODAY());
        //      IF ResourceLocation.FINDFIRST THEN BEGIN
        //        FilterString := ResourceLocation."Location Code";
        //        WHILE ResourceLocation.NEXT <> 0 DO BEGIN
        //           FilterString := FilterString + '|'  + ResourceLocation."Location Code";
        //        END;
        //        FldRef := RecRef.FIELD(67); // Location Filter
        //        FldRef.SETFILTER(FilterString);
        //      END;
        //    END;
        //  END;

        //  Database::Contact: BEGIN
        //    FldRef := RecRef.FIELD(5327201); // ACF Favorite User Filter
        //    FldRef.SETRANGE(Mgt.GetACFUser());
        //  END;

        //  Database::Customer: BEGIN
        //    FldRef := RecRef.FIELD(5327201); // ACF Favorite User Filter
        //    FldRef.SETRANGE(Mgt.GetACFUser());
        //  END;

        END;
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'SetBlobDownloadType', '', false, false)]
    local procedure SetBlobDownloadType(RecRef: RecordRef; FldRef: FieldRef; AppField: Record "ACF App Field"; VAR BlobDownloadType: Integer)
    var
        FileRepository: Record "ACF File Repository";
        DownloadTypeOption: Option "Synch.","Async.","WIFI/LAN only",,,,,,,,,,,,,,,,,,,,,,,,,,,,,,"Async. 4G+","Async. 3G+","Async. 2G+";
    begin

        // Set/overwrite how to load BLOB field values to the app
        // Use variable DownloadTypeOption to set valid values

        // Load data under 500KB directly
        if Mgt.GetBlobFieldLengthInKB(FldRef) < 500 then begin
            BlobDownloadType := DownloadTypeOption::"Synch.";
            exit;
        end;

        case RecRef.Number of
            Database::"ACF File Repository":
                begin
                    RecRef.SetTable(FileRepository);
                    BlobDownloadType := FileRepository."Download Type";
                end;
        // Example: Set picture field of table Item to load asynchronously when WIFI is available
        //Database::Item: begin
        //  IF FldRef.Number = 92 THEN
        //    BlobDownloadType := DownloadTypeOption::"Async. WIFI/LAN only";
        //end;
        end;
    end;

    [EventSubscriber(ObjectType::Codeunit, Codeunit::"ACF App Events", 'SetAsyncDownloadData', '', false, false)]
    local procedure SetAsyncDownloadData(RecRef: RecordRef; FldRef: FieldRef; AppField: Record "ACF App Field"; var DataBT: BigText)
    begin
        case RecRef.Number of
        // Example: Load image file from file system to be used as item picture on devices
        //Database::Item: begin
        //  RecRef.SetTable(Item);
        //  if FldRef.Number = 92 then begin
        //    Mgt.LoadFileToBigText('Path/to/files' + Item."No." + '.jpg', DataBT);
        //  end;
        //end;
        end;
    end;

    procedure BCFunction(FileRepository: Record "ACF File Repository")
    var
        SalesHeader: Record "Sales Header";
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
        ReleaseSalesDoc: Codeunit "Release Sales Document";
    begin
        // Example For Request From App To Call A Dynamics Function
        case FileRepository."Content Type" of
            //SH_REOPEN_REQ: Sales Header Should Not Be Reopen By Mobile Device. Sales Header Could Be Posted or Deleted
            'SH_REOPEN_REQ':
                begin
                    SalesHeader."Document Type" := "Sales Document Type".FromInteger(FileRepository."Document Type");
                    SalesHeader."No." := FileRepository.Code;
                    if TranslateSalesHeader(SalesHeader) then
                        ReleaseSalesDoc.PerformManualReopen(SalesHeader)
                    else
                        //Send Notification Sales Header Does Not Exists
                        SendNotificationToAnveoUser(Mgt.GetACFUser(), 36, FileRepository."Document Type", FileRepository.Code,
                                                  StrSubstNo(Text002, SalesHeader."Document Type", SalesHeader."No."));
                end;
            'SH:CANCELAPPR_REQ':
                begin
                    SalesHeader."Document Type" := "Sales Document Type".FromInteger(FileRepository."Document Type");
                    SalesHeader."No." := FileRepository.Code;
                    if TranslateSalesHeader(SalesHeader) then begin
                        if (SalesHeader.Status = SalesHeader.Status::Released) then
                            SendNotificationSHToSalesPer(SalesHeader, StrSubstNo(Text007, SalesHeader."Document Type", SalesHeader."No."))
                    end else
                        SendNotificationToAnveoUser(Mgt.GetACFUser(), 36, FileRepository."Document Type", FileRepository.Code,
                                                  StrSubstNo(Text002, SalesHeader."Document Type", SalesHeader."No."));
                    //Send Notification Sales Header Does Not Exists
                end;


        end;

        //Example For Other Notification In Integration In BC
        //Codeunit 414 - Release Sales Document
        //Function: PerformManualReopen
        //ACFAppEvents.SendNotificationSHToSalesPer(SalesHeader, STRSUBSTNO(Text81000, SalesHeader."Document Type", SalesHeader."No."));
    end;

    /***************** Notifications ***************************/
    // ***** IMPORTANT DISCLAIMER *****
    // Notifications can be send to your mobile deviced without the need of synchronization, just an online connection is required.
    // Anveo cannot guarantee the transfer to your mobile device because of several third party dependencies like required services
    // from Apple, Google, your mobile network provider and others.
    // This is why Anveo may discontinue this feature because of these dependencies at any time.

    procedure SendNotificationToAnveoUser(AnveoUserCode: Code[10]; LinkedTabletNo: Integer; LinkedType: Option; LinkedCode: Code[20]; MessageText: Text[1024])
    var
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
    begin
        //Send Notification To Anveo Mobile App Users
        if AnveoUser.Get(AnveoUserCode) then
            Mgt.SendNewLinkedNotification(AnveoUser."Anveo User", MessageText, true, LinkedTabletNo, LinkedType, LinkedCode, 0);
    end;

    procedure SendNotificationToSalesPer(SalesPersonCode: Code[20]; LinkedTabletNo: Integer; LinkedType: Option; LinkedCode: Code[20]; MessageText: Text[1024])
    var
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
    begin
        //Send Notification To Anveo Mobile App Users Based On Relationtype Salesperson
        AnveoUser.SetRange("Relation Type", AnveoUser."Relation Type"::"Salesperson/Purchaser");
        AnveoUser.SetRange("Relation Code", SalesPersonCode);
        AnveoUser.SetFilter("Main Menu Mobile App", '<>%1', '');
        if AnveoUser.FindFirst then
            repeat
                Mgt.SendNewLinkedNotification(AnveoUser."Anveo User", MessageText, true, LinkedTabletNo, LinkedType, LinkedCode, 0);
            until AnveoUser.Next = 0;
    end;

    procedure SendNotificationToCust(LinkedCode: Code[20]; MessageText: Text[1024])
    var
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
    begin
        //Send Notification To Anveo Mobile App Users Based On Relationtype Customer
        AnveoUser.SetRange("Relation Type", AnveoUser."Relation Type"::Customer);
        AnveoUser.SetRange("Relation Code", LinkedCode);
        AnveoUser.SetFilter("Main Menu Mobile App", '<>%1', '');
        if AnveoUser.FindFirst then
            repeat
                Mgt.SendNewLinkedNotification(AnveoUser."Anveo User", MessageText, true, Database::Customer, 0, LinkedCode, 0);
            until AnveoUser.Next = 0;
    end;

    procedure SendNotificationSHToSalesPer(SalesHeader: Record "Sales Header"; MessageText: Text[1024])
    var
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
    begin
        //Send Notification To Anveo Mobile App Users Based On Sales Header
        if SalesHeader."Salesperson Code" <> '' then begin
            AnveoUser.SetRange(AnveoUser."Relation Type", AnveoUser."Relation Type"::"Salesperson/Purchaser");
            AnveoUser.SetRange(AnveoUser."Relation Code", SalesHeader."Salesperson Code");
            AnveoUser.SetFilter("Main Menu Mobile App", '<>%1', '');
            if AnveoUser.FindFirst then
                repeat

                    Mgt.SendNewLinkedNotification(AnveoUser."Anveo User",
                      MessageText, true, Database::"Sales Header", SalesHeader."Document Type".AsInteger(), SalesHeader."No.", 0);
                until AnveoUser.Next = 0;
        end;
    end;

    procedure SendNotificationServOAToRes(ServiceOrderAllocation: Record "Service Order Allocation"; MessageText: Text[1024])
    var
        AnveoUser: Record "ACF Anveo User";
        Notification: Record "ACF Notification";
    begin
        //Send Notification To Anveo Mobile App Users Based On Service Line
        if ServiceOrderAllocation."Resource No." <> '' then begin
            AnveoUser.SetRange(AnveoUser."Relation Type", AnveoUser."Relation Type"::Resource);
            AnveoUser.SetRange(AnveoUser."Relation Code", ServiceOrderAllocation."Resource No.");
            AnveoUser.SetFilter("Main Menu Mobile App", '<>%1', '');
            if AnveoUser.FindFirst then
                repeat

                    Mgt.SendNewLinkedNotification(AnveoUser."Anveo User",
                      MessageText, true, Database::"Service Order Allocation",
                      ServiceOrderAllocation."Document Type".AsInteger(),
                      ServiceOrderAllocation."Document No.", 0);
                until AnveoUser.Next = 0;
        end;
    end;

    local procedure FldRefToText(FldRef: FieldRef): Text[100]
    begin
        // Used To Show Field Content In Log Files
        if (UpperCase(Format(FldRef.Class)) = 'NORMAL') and (UpperCase(Format(FldRef.Type)) <> 'BLOB') then
            exit(CopyStr(Format(FldRef.Value), 1, 100));
    end;

    local procedure CompareMobileAppRecord(xRecRef: RecordRef; RecRef: RecordRef; var AvailableFields: Record "Integer") RecordsMatch: Boolean
    var
        xFldRef: FieldRef;
        FldRef: FieldRef;
        FieldMatches: Boolean;
    begin
        // Internal Function That Compares All Fields Listed In (Temporary) AvailableFields Table Of Two RecordRefs
        RecordsMatch := true;
        if AvailableFields.FindFirst then
            repeat
                xFldRef := xRecRef.Field(AvailableFields.Number);
                FldRef := RecRef.Field(AvailableFields.Number);
                if (UpperCase(Format(FldRef.Class)) = 'NORMAL') and (Format(FldRef.Type) <> 'BLOB') then begin
                    FieldMatches := xFldRef.Value = FldRef.Value;
                    RecordsMatch := RecordsMatch and FieldMatches;
                    if not FieldMatches then
                        Mgt.Log('OnSetValue', StrSubstNo('Field %1 %2 is different. xDeviceRecRef:%3 RecRef:%4',
                          FldRef.Number, FldRef.Name, FldRefToText(xFldRef), FldRefToText(FldRef)));
                end;
            until (AvailableFields.Next = 0); //Improve Performance: OR (NOT RecordsMatch);
    end;

    /**** Temporary Rec Translation ****/

    local procedure TranslateSalesHeader(var SalesHeader: Record "Sales Header"): Boolean
    var
        TranslateRecRef: RecordRef;
    begin
        // Get Translated Sales Header To Perform Functions
        TranslateRecRef.GetTable(SalesHeader);
        if GetMatchedRecRef(TranslateRecRef) then
            TranslateRecRef.SetTable(SalesHeader);
        exit(SalesHeader.Get(SalesHeader."Document Type", SalesHeader."No."));
    end;

    local procedure TranslateServiceHeader(var ServiceHeader: Record "Service Header"): Boolean
    var
        TranslateRecRef: RecordRef;
    begin
        // Get Translated Sales Header To Perform Functions
        TranslateRecRef.GetTable(ServiceHeader);
        if GetMatchedRecRef(TranslateRecRef) then
            TranslateRecRef.SetTable(ServiceHeader);
        exit(ServiceHeader.Get(ServiceHeader."Document Type", ServiceHeader."No."));
    end;

    procedure TranslatePurchaseHeader(var PurchaseHeader: Record "Purchase Header"): Boolean
    var
        TranslateRecRef: RecordRef;
    begin

        // Get Translated Purchase Header To Perform Functions
        TranslateRecRef.GetTable(PurchaseHeader);
        if GetMatchedRecRef(TranslateRecRef) then
            TranslateRecRef.SetTable(PurchaseHeader);
        exit(PurchaseHeader.Get(PurchaseHeader."Document Type", PurchaseHeader."No."));
    end;

    /**** Temporary PK Translations ****/

    procedure TranslateSalesHeaderNo(DocumentType: Integer; IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record "Sales Header";
        TranslateRecRef: RecordRef;
    begin
        //Translate Sales Header Code From Table Sales Header
        if IncomingValue = '' then exit('');
        TranslateRecord."Document Type" := "Sales Document Type".FromInteger(DocumentType);
        ;
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateServiceHeaderNo(DocumentType: Integer; IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record "Service Header";
        TranslateRecRef: RecordRef;
    begin
        //Translate Service Header Code From Table Service Header
        if IncomingValue = '' then exit('');
        TranslateRecord."Document Type" := "Sales Document Type".FromInteger(DocumentType);
        ;
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslatePurchaseHeaderNo(DocumentType: Integer; IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record "Purchase Header";
        TranslateRecRef: RecordRef;
    begin
        //Translate Purchase Header Code From Table Purchase Header
        if IncomingValue = '' then exit('');
        TranslateRecord."Document Type" := "Sales Document Type".FromInteger(DocumentType);
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateCustomerNo(IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record Customer;
        TranslateRecRef: RecordRef;
    begin
        //Translate Customer No. From Table Customer
        if IncomingValue = '' then exit('');
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateVendorNo(IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record Vendor;
        TranslateRecRef: RecordRef;
    begin
        //Translate Customer No. From Table Customer
        if IncomingValue = '' then exit('');
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateContactNo(IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record Contact;
        TranslateRecRef: RecordRef;
    begin
        //Translate Contact No. From Table Contact
        if IncomingValue = '' then exit('');
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateOppNo(IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record Opportunity;
        TranslateRecRef: RecordRef;
    begin
        //Translate Opp No. From Table Opportunity
        if IncomingValue = '' then exit('');
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateShipAddCode(IncomingValue: Code[20]; CustomerNo: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record "Ship-to Address";
        TranslateRecRef: RecordRef;
        ShipToAdress: Record "Ship-to Address";
    begin
        //Translate Opp No. From Table Opportunity
        if IncomingValue = '' then exit('');
        TranslateRecord.Code := IncomingValue;
        TranslateRecord."Customer No." := CustomerNo;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord.Code);
    end;

    procedure TranslateServiceItemNo(IncomingValue: Code[20]) TranslatedValue: Code[20]
    var
        TranslateRecord: Record "Service Item";
        TranslateRecRef: RecordRef;
    begin
        //Translate Service Item No. from Table Service Item
        if IncomingValue = '' then exit('');
        TranslateRecord."No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."No.");
    end;

    procedure TranslateServiceItemLineNo(IncomingValue: Integer) TranslatedValue: Integer
    var
        TranslateRecord: Record "Service Item Line";
        TranslateRecRef: RecordRef;
    begin
        //Translate Service Item Line No. from Table Service Item Line
        if IncomingValue = 0 then exit(0);
        TranslateRecord."Line No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."Line No.");
    end;

    procedure TranslateMeetingReportNo(IncomingValue: Integer) TranslatedValue: Integer
    var
        TranslateRecord: Record "ACF Meeting Report";
        TranslateRecRef: RecordRef;
    begin
        //Translate Service Header Code From Table Service Header
        if IncomingValue = 0 then exit(0);
        TranslateRecord."Entry No." := IncomingValue;
        TranslateRecRef.GetTable(TranslateRecord);
        if GetMatchedRecRef(TranslateRecRef) then TranslateRecRef.SetTable(TranslateRecord);
        exit(TranslateRecord."Entry No.");
    end;

    local procedure GetMatchedRecRef(var pRecRef: RecordRef) FoundMatching: Boolean
    var
        MobileNoMatching: Record "ACF Mobile No. Matching";
    begin
        // Get A Matching For Mobile No. Series Management
        Mgt.Log('GetMatchedRecRef', 'BEGIN From PK:' + Format(pRecRef.RecordId));
        MobileNoMatching.SetRange("Anveo User Device ID", Mgt.GetAUDID);
        MobileNoMatching.SetRange("From PK", pRecRef.RecordId);
        if MobileNoMatching.FindFirst then begin
            pRecRef.Get(MobileNoMatching."To PK");
            Mgt.Log('GetMatchedRecRef', 'END To PK:' + Format(MobileNoMatching."To PK"));
            exit(true);
        end;
        Mgt.Log('GetMatchedRecRef', 'END No matching found.');
    end;

    local procedure PrintReport(FileRepository: Record "ACF File Repository")
    var
        PDFSaved: Boolean;
        //ExcelSaved: Boolean;
        OutStr: OutStream;
        RecRef: RecordRef;
        "--Records--": Integer;
        SalesCrMemoHeader: Record "Sales Cr.Memo Header";
        ServiceHeader: Record "Service Header";
        ServiceItemLine: Record "Service Item Line";
        SalesHeader: Record "Sales Header";
        PurchaseHeader: Record "Purchase Header";
        SalesShipmentHeader: Record "Sales Shipment Header";
        SalesInvoiceHeader: Record "Sales Invoice Header";
        ServiceShipmentHeader: Record "Service Shipment Header";
        ReportSelections: Record "Report Selections";
        PurchaseReceiptHeader: Record "Purch. Rcpt. Header";
        PurchaseInvoiceHeader: Record "Purch. Inv. Header";
        PurchaseCrMemoHeader: Record "Purch. Cr. Memo Hdr.";
        "--Reports--": Integer;
        ServiceItemWorksheet: Report "Service Item Worksheet";
        SalesCreditMemo: Report "Standard Sales - Credit Memo";
        SalesQuote: Report "Standard Sales - Quote";
        ServiceOrder: Report "Service Order";
        OrderConfirmation: Report "Standard Sales - Order Conf.";
        SalesShipment: Report "Sales - Shipment";
        SalesInvoice: Report "Standard Sales - Invoice";
        PurchaseOrderConfirmation: Report "Order";
        PurchaseQuote: Report "Purchase - Quote";
        PurchaseInvoice: Report "Purchase - Invoice";
        PurchaseReceipt: Report "Purchase - Receipt";
        PurchaseCreditMemo: Report "Purchase - Credit Memo";
    begin
        Mgt.Log('PrintReport', 'BEGIN');
        FileRepository.Data.CreateOutStream(OutStr);

        case FileRepository."Table No." of
            // Example: Print Order Confirmation as PDF file
            Database::"Sales Header":
                begin
                    if SalesHeader.Get(FileRepository."Document Type", FileRepository.Code) then begin
                        SalesHeader.SetRecFilter;
                        RecRef.GetTable(SalesHeader);
                        if FileRepository."Content Type" = 'ORDERCONFIRMATION' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"S.Order");
                        end;
                        if FileRepository."Content Type" = 'QUOTE' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"S.Quote");
                        end;
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;

            Database::"Purchase Header":
                begin
                    if PurchaseHeader.Get(FileRepository."Document Type", FileRepository.Code) then begin
                        PurchaseHeader.SetRecFilter;
                        RecRef.GetTable(PurchaseHeader);
                        if FileRepository."Content Type" = 'ORDERCONFIRMATION' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"P.Order");
                        end;
                        if FileRepository."Content Type" = 'QUOTE' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"P.Quote");
                        end;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;

            Database::"Service Header":
                begin
                    if ServiceHeader.Get(FileRepository."Document Type", FileRepository.Code) then begin
                        ServiceHeader.SetRecFilter;
                        RecRef.SetTable(ServiceHeader);
                        if FileRepository."Content Type" = 'SERVICEORDER' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"SM.Order");
                        end;
                        if FileRepository."Content Type" = 'SERVICEQUOTE' then begin
                            ReportSelections.SetRange(Usage, ReportSelections.Usage::"SM.Quote");
                        end;
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;
            Database::"Service Item Line":
                begin
                    if ServiceItemLine.Get(FileRepository."Document Type", FileRepository.Code, FileRepository."Line No.") then begin
                        ServiceItemLine.SetRecFilter;
                        ServiceItemWorksheet.SetTableView(ServiceItemLine);
                        PDFSaved := ServiceItemWorksheet.SaveAs('', ReportFormat::Pdf, OutStr);
                    end;
                end;

            // Example: Print Sales - Shipment as PDF file
            Database::"Sales Shipment Header":
                begin
                    if SalesShipmentHeader.Get(FileRepository.Code) then begin
                        SalesShipmentHeader.SetRecFilter;
                        RecRef.SetTable(SalesShipmentHeader);
                        ReportSelections.SetRange(Usage, ReportSelections.Usage::"S.Shipment");
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;
            //Example: Print Sales - Invoice as PDF file
            Database::"Sales Invoice Header":
                begin
                    if SalesInvoiceHeader.Get(FileRepository.Code) then begin
                        SalesInvoiceHeader.SetRecFilter;
                        RecRef.SetTable(SalesInvoiceHeader);
                        ReportSelections.SetRange(Usage, ReportSelections.Usage::"S.Invoice");
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;

            Database::"Sales Cr.Memo Header":
                begin
                    if SalesCrMemoHeader.Get(FileRepository.Code) then begin
                        SalesCrMemoHeader.SetRecFilter;
                        RecRef.GetTable(SalesCrMemoHeader);
                        ReportSelections.SetRange(Usage, ReportSelections.Usage::"S.Cr.Memo");
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;

            Database::"Service Shipment Header":
                begin
                    if ServiceShipmentHeader.Get(FileRepository.Code) then begin
                        ServiceShipmentHeader.SetRecFilter;
                        RecRef.GetTable(ServiceShipmentHeader);
                        ReportSelections.SetRange(Usage, ReportSelections.Usage::"SM.Shipment");
                        ReportSelections.FindFirst;
                        PDFSaved := REPORT.SaveAs(ReportSelections."Report ID", '', ReportFormat::Pdf, OutStr, RecRef);
                    end;
                end;

            // Example: Print Purchase - Receipt as PDF file
            Database::"Purch. Rcpt. Header":
                begin
                    if PurchaseReceiptHeader.Get(FileRepository.Code) then begin
                        PurchaseReceiptHeader.SetRecFilter;
                        PurchaseReceipt.SetTableView(PurchaseReceiptHeader);
                        PDFSaved := PurchaseReceipt.SaveAs('', ReportFormat::Pdf, OutStr);
                    end;
                end;
            //Example: Print Purchase - Invoice as PDF file
            Database::"Purch. Inv. Header":
                begin
                    if PurchaseInvoiceHeader.Get(FileRepository.Code) then begin
                        PurchaseInvoiceHeader.SetRecFilter;
                        PurchaseInvoice.SetTableView(PurchaseInvoiceHeader);
                        PDFSaved := PurchaseInvoice.SaveAs('', ReportFormat::Pdf, OutStr);
                    end;
                end;
            Database::"Purch. Cr. Memo Hdr.":
                begin
                    if PurchaseCrMemoHeader.Get(FileRepository.Code) then begin
                        PurchaseCrMemoHeader.SetRecFilter;
                        PurchaseCreditMemo.SetTableView(PurchaseCrMemoHeader);
                        PDFSaved := PurchaseCreditMemo.SaveAs('', ReportFormat::Pdf, OutStr);
                    end;
                end;
        end;

        if PDFSaved then begin
            FileRepository.Type := FileRepository.Type::Document;
            FileRepository.Filename := FileRepository."Content Type" + '_' + Format(FileRepository.Code) + '.pdf';
            FileRepository.Modify(true);
        end;

        Mgt.Log('PrintReport', StrSubstNo('END PDFSaved:%1', PDFSaved));
    end;

}